#include <Python.h>
#include <pvm3.h>
#include <string.h>
#include <stdlib.h>


/* Useful definitions for this module */

static PyObject * pypvm_module;
static PyObject * pypvm_dictionary;

static PyObject 
  *PvmBadParamException, *PvmMismatchException,	 *PvmOverflowException,	 
  *PvmNoDataException ,	 *PvmNoHostException,	 *PvmNoFileException,	 
  *PvmNoMemException ,	 *PvmBadMsgException,	 *PvmSysErrException,	 
  *PvmNoBufException ,	 *PvmNoSuchBufException, *PvmNullGroupException,
  *PvmDupGroupException, *PvmNoGroupException,   *PvmNotInGroupException ,
  *PvmNoInstException ,	 *PvmHostFailException,  *PvmNoParentException ,	 
  *PvmNotImplException , *PvmDSysErrException,	 *PvmBadVersionException,	 
  *PvmOutOfResException, *PvmDupHostException,	 *PvmCantStartException,	 
  *PvmAlreadyException , *PvmNoTaskException,	 *PvmNoEntryException,	 
  *PvmDupEntryException;

static PyObject 
  *PvmBadParamNumber, *PvmMismatchNumber,	 *PvmOverflowNumber,	 
  *PvmNoDataNumber ,	 *PvmNoHostNumber,	 *PvmNoFileNumber,	 
  *PvmNoMemNumber ,	 *PvmBadMsgNumber,	 *PvmSysErrNumber,	 
  *PvmNoBufNumber ,	 *PvmNoSuchBufNumber, *PvmNullGroupNumber,
  *PvmDupGroupNumber, *PvmNoGroupNumber,   *PvmNotInGroupNumber ,
  *PvmNoInstNumber ,	 *PvmHostFailNumber,  *PvmNoParentNumber ,	 
  *PvmNotImplNumber , *PvmDSysErrNumber,	 *PvmBadVersionNumber,	 
  *PvmOutOfResNumber, *PvmDupHostNumber,	 *PvmCantStartNumber,	 
  *PvmAlreadyNumber , *PvmNoTaskNumber,	 *PvmNoEntryNumber,	 
  *PvmDupEntryNumber;

/* There are lots of functions that may want to return nothing,  or
   throw an exception if their corresponding PVM call returned < 0.
   We define a macro `return_none' to handle this.  Its argument is
   the return result from the PVM call.  (Which will probably be
   a variable name. */

#define return_none(x) \
  if (was_error(x)) return NULL; \
  Py_INCREF(Py_None); \
  return Py_None;

/* PVM often returns negative numbers to mean some kind of error.
   Python can throw exceptions in these cases,  so that's what
   we do.  (We use set object with the exception as the string,
   and the data as the PVM return value.  Maybe it should be
   the other way around.  Anyway,  it's kind of cute. */


static int was_error(int info) {
  if (info > 0) return 0;
  switch (info) {
  case PvmOk: return 0;
#define error_case(x) case x : \
    PyErr_SetObject(x ## Exception,x ## Number); \
    return 1;
  error_case(PvmBadParam); error_case(PvmMismatch);  error_case(PvmOverflow);
  error_case(PvmNoData);  error_case(PvmNoHost);  error_case(PvmNoFile);
  error_case(PvmNoMem);  error_case(PvmBadMsg);  error_case(PvmSysErr);
  error_case(PvmNoBuf); error_case(PvmNoSuchBuf);  error_case(PvmNullGroup);
  error_case(PvmDupGroup);  error_case(PvmNoGroup);  error_case(PvmNotInGroup);
  error_case(PvmNoInst);  error_case(PvmHostFail);  error_case(PvmNoParent);
  error_case(PvmNotImpl);  error_case(PvmDSysErr);  error_case(PvmBadVersion);
  error_case(PvmOutOfRes);  error_case(PvmDupHost);  error_case(PvmCantStart);
  error_case(PvmAlready);  error_case(PvmNoTask);  error_case(PvmNoEntry);
  error_case(PvmDupEntry);
  }
}


static PyObject * pypvm_hostinfo (PyObject *self, PyObject *args) {
  struct pvmhostinfo *hinfo;
  int i;
  PyObject *list;
  int nhost;
  int narch;
  int info;

  PyObject *pvmd_tid, *hostname, *architecture, *speed;
  PyObject *tuple;

  info = pvm_config(&nhost,&narch,&hinfo);
  /* I should check to see if info == PvmSysErr,  and raise an exception if it is. */
  if (was_error(info)) return NULL;
  list = PyList_New(nhost);

  for (i = 0; i < nhost; i++) {
    pvmd_tid = PyInt_FromLong(hinfo[i].hi_tid);
    hostname = PyString_FromString(hinfo[i].hi_name);
    architecture = PyString_FromString(hinfo[i].hi_arch);
    speed = PyInt_FromLong(hinfo[i].hi_speed);
    
    tuple = PyTuple_New(4);
    PyTuple_SetItem(tuple,0,pvmd_tid);
    PyTuple_SetItem(tuple,1,hostname);
    PyTuple_SetItem(tuple,2,architecture); 
    PyTuple_SetItem(tuple,3,speed);
    PyList_SetItem(list,i,tuple);
  }
  return list;
}


static PyObject * pypvm_tasks (PyObject* self,PyObject* args) {
  struct pvmtaskinfo * taskinfo;
  int where=0;   /* If they don't provide it, they probably meant everything */
  int ntasks;
  int info;
  int i;

  PyObject * list, *tuple;
  
  if (!PyArg_ParseTuple(args,"|i",&where))  return NULL;

  info = pvm_tasks(where,&ntasks,&taskinfo);
  if (was_error(info)) return NULL;
  
  list = PyList_New(ntasks);
  for (i=0;i<ntasks;i++) {
    tuple = Py_BuildValue("(iiiisi)",
			  taskinfo[i].ti_tid,
			  taskinfo[i].ti_ptid,
			  taskinfo[i].ti_host,
			  taskinfo[i].ti_flag,
			  taskinfo[i].ti_a_out,
			  taskinfo[i].ti_pid);
    PyList_SetItem(list,i,tuple);
  }

  return list;
}


/* -------------------------------------------------------------------------- */
/* The following functions were originally created by swig, but I've modified them all */
/* -------------------------------------------------------------------------- */

static PyObject *pypvm_archcode(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;
    char * arch;

    self = self;
    if(!PyArg_ParseTuple(args,"s:pvm_archcode",&arch)) 
        return NULL;
    result = (int )pvm_archcode(arch);
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}

static PyObject *pypvm_barrier(PyObject *self, PyObject *args) {
    int  info;
    char * group;
    int  count;

    self = self;
    if(!PyArg_ParseTuple(args,"si:pvm_barrier",&group,&count)) 
        return NULL;
    info = pvm_barrier(group,count);
    return_none(info);
}

static PyObject *pypvm_bcast(PyObject *self, PyObject *args) {
    int  info;
    char * group;
    int  msgtag;

    self = self;
    if(!PyArg_ParseTuple(args,"si:pvm_bcast",&group,&msgtag)) 
        return NULL;
    info = pvm_bcast(group,msgtag);
    return_none(info);
}

static PyObject *pypvm_bufinfo(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  info;
    int  bufid;
    int  bytes, msgtag, source_tid;
    

    if(!PyArg_ParseTuple(args,"i:pvm_bufinfo",&bufid)) 
        return NULL;
    info = pvm_bufinfo(bufid,&bytes,&msgtag,&source_tid);
    if (was_error(info)) return NULL;
    resultobj = Py_BuildValue("(iii)",bytes,msgtag,source_tid);

    return resultobj;
}

static PyObject *pypvm_delete(PyObject *self, PyObject *args) {
    int  cc;
    char * name;
    int  index;

    if(!PyArg_ParseTuple(args,"si:pvm_delete",&name,&index)) 
        return NULL;
    cc = pvm_delete(name,index);
    return_none(cc);
}

/*
static PyObject *pypvm_delhosts(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;
    char ** arg0;
    int  arg1;
    int * arg2;
    char * argc0 = 0;
    char * argc2 = 0;

    self = self;
    if(!PyArg_ParseTuple(args,"sis:pvm_delhosts",&argc0,&arg1,&argc2)) 
        return NULL;
    if (argc0) {
        if (SWIG_GetPtr(argc0,(void **) &arg0,"_char_pp")) {
            PyErr_SetString(PyExc_TypeError,"Type error in argument 1 of pvm_delhosts. Expected _char_pp.");
        return NULL;
        }
    }
    if (argc2) {
        if (SWIG_GetPtr(argc2,(void **) &arg2,"_int_p")) {
            PyErr_SetString(PyExc_TypeError,"Type error in argument 3 of pvm_delhosts. Expected _int_p.");
        return NULL;
        }
    }
    result = (int )pvm_delhosts(arg0,arg1,arg2);
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}

*/

static PyObject *pypvm_exit(PyObject *self, PyObject *args) {
    int  result;

    result = pvm_exit();
    return_none(result);
}

static PyObject *pypvm_export(PyObject *self, PyObject *args) {
  int  cc;
  char * name;

  if(!PyArg_ParseTuple(args,"s:pvm_export",&name)) 
    return NULL;
  cc = pvm_export(name);
  return_none(cc);
}

static PyObject *pypvm_freebuf(PyObject *self, PyObject *args) {
    int  info;
    int  bufid;

    if(!PyArg_ParseTuple(args,"i:pvm_freebuf",&bufid)) 
        return NULL;
    info = pvm_freebuf(bufid);
    return_none(info);
}

/*
PyObject * pvm_gather(PyObject * self, PyObject * args) {
}
*/

/* This next function should return Python file objects.
   Better still,  we should have a python select wrapper as well. */

/*
static PyObject *pypvm_getfds(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;
    int ** arg0;
    char * argc0 = 0;

    self = self;
    if(!PyArg_ParseTuple(args,"s:pvm_getfds",&argc0)) 
        return NULL;
    if (argc0) {
        if (SWIG_GetPtr(argc0,(void **) &arg0,"_int_pp")) {
            PyErr_SetString(PyExc_TypeError,"Type error in argument 1 of pvm_getfds. Expected _int_pp.");
        return NULL;
        }
    }
    result = (int )pvm_getfds(arg0);
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}
*/


static PyObject *pypvm_getinst(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  inum;
    char * group;
    int  tid;

    if(!PyArg_ParseTuple(args,"si:pvm_getinst",&group,&tid)) 
        return NULL;
    inum= pvm_getinst(group,tid);
    if (was_error(inum)) return NULL;
    resultobj = Py_BuildValue("i",inum);
    return resultobj;
}

static PyObject *pypvm_getmwid(PyObject *self, PyObject *args) {
    int  info;
    int  bufid;

    self = self;
    if(!PyArg_ParseTuple(args,"i:pvm_getmwid",&bufid))  return NULL;
    info = pvm_getmwid(bufid);
    return_none(info);
}

/*
static PyObject *pypvm_getopt(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;
    int  arg0;

    self = self;
    if(!PyArg_ParseTuple(args,"i:pvm_getopt",&arg0)) 
        return NULL;
    result = (int )pvm_getopt(arg0);
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}
*/


static PyObject *pypvm_getrbuf(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;

    self = self;
    if(!PyArg_ParseTuple(args,":pvm_getrbuf")) 
        return NULL;
    result = (int )pvm_getrbuf();
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}

static PyObject *pypvm_getsbuf(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;

    self = self;
    if(!PyArg_ParseTuple(args,":pvm_getsbuf")) 
        return NULL;
    result = (int )pvm_getsbuf();
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}

static PyObject *pypvm_gettid(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  tid;
    char * group;
    int  inum;

    if(!PyArg_ParseTuple(args,"si:pvm_gettid",&group,&inum)) return NULL;
    tid = pvm_gettid(group,inum);
    if (was_error(tid)) return NULL;

    resultobj = Py_BuildValue("i",tid);
    return resultobj;
}

static PyObject *pypvm_gsize(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;
    char * group;

    self = self;
    if(!PyArg_ParseTuple(args,"s:pvm_gsize",&group)) return NULL;
    result = (int )pvm_gsize(group);
    if (was_error(result)) return NULL;
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}

static PyObject *pypvm_halt(PyObject *self, PyObject *args) {
    int  result;

    result = pvm_halt();
    return_none(result);
}

static PyObject *pypvm_initsend(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  bufid;
    int  encoding;

    encoding = PvmDataDefault;  /* if they don't provide an argument, then use XDR */
    if(!PyArg_ParseTuple(args,"|i",&encoding)) return NULL;

    bufid = pvm_initsend(encoding);
    if (was_error(bufid)) return NULL;
    resultobj = Py_BuildValue("i",bufid);
    return resultobj;
}

static PyObject *pypvm_insert(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  cc;
    char * name;
    int  index;
    int  data;

    if(!PyArg_ParseTuple(args,"sii:pvm_insert",&name,&index,&data)) return NULL;
    cc = pvm_insert(name,index,data);
    if (was_error(cc)) return NULL;
    resultobj = Py_BuildValue("i",cc);
    return resultobj;
}

static PyObject *pypvm_joingroup(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  inum;
    char * group;

    self = self;
    if(!PyArg_ParseTuple(args,"s:pvm_joingroup",&group)) return NULL;
    inum= pvm_joingroup(group);
    if (was_error(inum)) return NULL;
    resultobj = Py_BuildValue("i",inum);
    return resultobj;
}

static PyObject *pypvm_kill(PyObject *self, PyObject *args) {
    int  info;
    int  tid;

    if(!PyArg_ParseTuple(args,"i:pvm_kill",&tid)) return NULL;
    info = pvm_kill(tid);
    return_none(info);
}

static PyObject *pypvm_lookup(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  cc;
    char * name;
    int  index;
    int data;

    if(!PyArg_ParseTuple(args,"si:pvm_lookup",&name,&index)) return NULL;
    cc = pvm_lookup(name,index,&data);
    if (was_error(cc)) return NULL;
    resultobj = Py_BuildValue("i",cc);
    return resultobj;
}

static PyObject *pypvm_lvgroup(PyObject *self, PyObject *args) {
    int  info;
    char * group;

    self = self;
    if(!PyArg_ParseTuple(args,"s:pvm_lvgroup",&group)) 
        return NULL;
    info = pvm_lvgroup(group);
    return_none(info);
}

/*
static PyObject *pypvm_mcast(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;
    int * arg0;
    int  arg1;
    int  arg2;
    char * argc0 = 0;

    self = self;
    if(!PyArg_ParseTuple(args,"sii:pvm_mcast",&argc0,&arg1,&arg2)) 
        return NULL;
    if (argc0) {
        if (SWIG_GetPtr(argc0,(void **) &arg0,"_int_p")) {
            PyErr_SetString(PyExc_TypeError,"Type error in argument 1 of pvm_mcast. Expected _int_p.");
        return NULL;
        }
    }
    result = (int )pvm_mcast(arg0,arg1,arg2);
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}
*/

static PyObject *pypvm_mkbuf(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  bufid;
    int  encoding;

    encoding = PvmDataDefault;   /* if they don't provide an encoding,  they mean XDR */
    if(!PyArg_ParseTuple(args,"|i",&encoding)) return NULL;
    bufid= pvm_mkbuf(encoding);
    if (was_error(bufid)) return NULL;
    resultobj = Py_BuildValue("i",bufid);
return resultobj;
}

static PyObject *pypvm_mstat(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  mstat;
    char * host;

    if(!PyArg_ParseTuple(args,"s:pvm_mstat",&host)) return NULL;
    mstat = pvm_mstat(host);
    if (was_error(mstat)) return NULL;
    resultobj = Py_BuildValue("i",mstat);
    return resultobj;
}

static PyObject *pypvm_mytid(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int tid = pvm_mytid();
    if (was_error(tid)) return NULL;
    resultobj = Py_BuildValue("i",tid);
    return resultobj;
}

static PyObject *pypvm_nrecv(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  bufid;
    int  tid=-1;
    int  msgtag=-1;

    self = self;
    if(!PyArg_ParseTuple(args,"|ii:pvm_nrecv",&tid,&msgtag)) return NULL;
    bufid = pvm_nrecv(tid,msgtag);
    if (was_error(bufid)) return NULL;
    resultobj = Py_BuildValue("i",bufid);
    return resultobj;
}

static PyObject *pypvm_parent(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  tid;

    tid =pvm_parent();
    if (was_error(tid)) return NULL;
    resultobj = Py_BuildValue("i",tid);
    return resultobj;
}

/* Hmm... should this function be used as the detail field for raised exceptions? */
static PyObject *pypvm_perror(PyObject *self, PyObject *args) {
    char * msg = "";

    if(!PyArg_ParseTuple(args,"s:pvm_perror",&msg)) {   return NULL;  }
    pvm_perror(msg);
    Py_INCREF(Py_None);
    return Py_None;
}

/* this is a utility function used by pypvm */
static PyObject * pack_one_object(PyObject *object) {
  int info;
  char * string;
  long integer;
  double real;
  if (PyString_Check(object)) {
    string = PyString_AsString(object);
    info = pvm_pkstr(string);
  } else if (PyInt_Check(object)) {
    integer = PyInt_AsLong(object);
    info = pvm_pklong(&integer,1,1);
  } else if (PyFloat_Check(object)) {
    real = PyFloat_AsDouble(object);
    info = pvm_pkdouble(&real,1,1);
  } else {
    fprintf(stderr,"Gotta make this an exception as well\n");
    exit(1);
    return NULL;
  }
  return_none(info);
}

/* The Python function defined by pypvm_pack takes one argument,
   hopefully a list. */
static PyObject *pypvm_pack(PyObject *self,PyObject *args) {
  PyObject * theList, *thisElement, *returnObject;
  int size, index;
  
  
  if (!PyTuple_Check(args)) { fprintf(stderr,"Seriously confused.\n"); exit(1); }
  size  = PyTuple_Size(args);
  if (size < 1) { fprintf(stderr,"Gotta make this into an exception\n"); exit(1); }
  theList = PyTuple_GetItem(args,0);
  if (!PyList_Check(theList)) { 
    /* It's not a list... perhaps they just want to pack one value.  Fair enough.
       Of course,  calling the thing they gave us "theList" is terribly confusing. */
    return (pack_one_object(theList));
  }

  /* Good.  Our first argument is a list.  Let's go through it item by item. */
  size = PyList_Size(theList);
  for (index=0;index<size;index++) {
    thisElement = PyList_GetItem(theList,index);
    returnObject = pack_one_object(thisElement);
    if (returnObject == NULL) return NULL;  /* if we couldn't pack it,  error out */
    Py_DECREF(Py_None);  /* because return_none inside pack_one_object does a Py_INCREF */
  }
  Py_INCREF(Py_None);  /* we are returning Py_None, so re-increment it back up again! */
  return Py_None;
}

static PyObject *pypvm_probe(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  bufid;
    int  tid;
    int  msgtag;

    self = self;
    if(!PyArg_ParseTuple(args,"ii:pvm_nrecv",&tid,&msgtag)) {
      /* Hmm.  What I really want to do here is have -1 as a default for each. How? */
      return NULL;
    }
    bufid = pvm_probe(tid,msgtag);
    if (was_error(bufid)) return NULL;
    resultobj = Py_BuildValue("i",bufid);
    return resultobj;
}

static PyObject *pypvm_pstat(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  status;
    int  tid;

    if(!PyArg_ParseTuple(args,"i:pvm_pstat",&tid)) 
        return NULL;
    status = pvm_pstat(tid);
    resultobj = Py_BuildValue("i",status);
    return resultobj;
}

static PyObject *pypvm_recv(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  bufid;
    int  tid=-1;
    int  msgtag=-1;

    if(!PyArg_ParseTuple(args,"|ii:pvm_recv",&tid,&msgtag))  return NULL;
    bufid = pvm_recv(tid,msgtag);
    if (was_error(bufid)) return NULL;
    resultobj = Py_BuildValue("i",bufid);
    return resultobj;
}

static PyObject *pypvm_reg_tasker(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  cc;

    cc = pvm_reg_tasker();
    return_none(cc); 
}

static PyObject *pypvm_send(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  info;
    int  tid;
    int  msgtag;

    if(!PyArg_ParseTuple(args,"ii:pvm_send",&tid,&msgtag)) return NULL;
    info= pvm_send(tid,msgtag);
    return_none(info);
}

static PyObject *pypvm_sendsig(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  info;
    int  tid;
    int  signum;

    if(!PyArg_ParseTuple(args,"ii:pvm_sendsig",&tid,&signum)) return NULL;
    info = pvm_sendsig(tid,signum);
    return_none(info);
}

static PyObject *pypvm_setmwid(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  info;
    int  bufid;
    int  waitid;

    if(!PyArg_ParseTuple(args,"ii:pvm_setmwid",&bufid,&waitid)) 
        return NULL;
    info = pvm_setmwid(bufid,waitid);
    return_none(info);
}

/* Needs to be wrapped into a dictionary */
/*
static PyObject *pypvm_setopt(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;
    int  arg0;
    int  arg1;

    self = self;
    if(!PyArg_ParseTuple(args,"ii:pvm_setopt",&arg0,&arg1)) 
        return NULL;
    result = (int )pvm_setopt(arg0,arg1);
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}
*/


static PyObject *pypvm_setrbuf(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  oldbufid;
    int  bufid;

    if(!PyArg_ParseTuple(args,"i:pvm_setrbuf",&bufid)) 
        return NULL;
    oldbufid = pvm_setrbuf(bufid);
    if (was_error(oldbufid)) return NULL;
    resultobj = Py_BuildValue("i",oldbufid);
    return resultobj;
}

static PyObject *pypvm_setsbuf(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  oldbufid;
    int  bufid;

    if(!PyArg_ParseTuple(args,"i:pvm_setrbuf",&bufid)) 
        return NULL;
    oldbufid = pvm_setsbuf(bufid);
    if (was_error(oldbufid)) return NULL;
    resultobj = Py_BuildValue("i",oldbufid);
    return resultobj;
}

static PyObject *pypvm_spawn(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  numt;
    char *task;
    PyObject *argvObject;
    char **argv;
    int flag;  
    char * where;
    int ntasks;
    int * tids;
    int size,index;

    if (!PyArg_ParseTuple(args,"sOisi",&task,&argvObject,&flag,&where,&ntasks)) {
      /* Hmm... I'd much rather have flags=default as default */
      return NULL;
    }

    size = PyList_Size(argvObject);
    argv = calloc(size+1,sizeof(char *));
    for (index=0;index<size;index++) {
      argv[index]=PyString_AsString(PyList_GetItem(argvObject,index));
    }
    argv[size]=NULL;

    tids=calloc(ntasks,sizeof(int));
    numt = pvm_spawn(task,argv,flag,where,ntasks,tids);
    free(argv);
    if (was_error(numt)) {
      free(tids);
      return NULL;
    }
    
    resultobj = PyList_New(ntasks);
    for (index=0;index<ntasks;index++) {
      PyList_SetItem(resultobj,index,PyInt_FromLong((long) tids[index]));
    }

    free(tids);
    return resultobj;
}

/* Is this one relevant? */
/*
static PyObject *pypvm_start_pvmd(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;
    int  arg0;
    char ** arg1;
    int  arg2;
    char * argc1 = 0;

    self = self;
    if(!PyArg_ParseTuple(args,"isi:pvm_start_pvmd",&arg0,&argc1,&arg2)) 
        return NULL;
    if (argc1) {
        if (SWIG_GetPtr(argc1,(void **) &arg1,"_char_pp")) {
            PyErr_SetString(PyExc_TypeError,"Type error in argument 2 of pvm_start_pvmd. Expected _char_pp.");
        return NULL;
        }
    }
    result = (int )pvm_start_pvmd(arg0,arg1,arg2);
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}
*/

static PyObject *pypvm_tidtohost(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  dtid;
    int  tid;

    if(!PyArg_ParseTuple(args,"i:pvm_tidtohost",&tid)) return NULL;
    dtid = pvm_tidtohost(tid);
    if (was_error(dtid)) return NULL;
    resultobj = Py_BuildValue("i",dtid);
    return resultobj;
}

/* This is urgently needed as well */
/*
static PyObject *pypvm_trecv(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    int  result;
    int  arg0;
    int  arg1;
    struct timeval * arg2;
    char * argc2 = 0;

    self = self;
    if(!PyArg_ParseTuple(args,"iis:pvm_trecv",&arg0,&arg1,&argc2)) 
        return NULL;
    if (argc2) {
        if (SWIG_GetPtr(argc2,(void **) &arg2,"_struct_timeval_p")) {
            PyErr_SetString(PyExc_TypeError,"Type error in argument 3 of pvm_trecv. Expected _struct_timeval_p.");
        return NULL;
        }
    }
    result = (int )pvm_trecv(arg0,arg1,arg2);
    resultobj = Py_BuildValue("i",result);
    return resultobj;
}
*/

static PyObject *pypvm_unexport(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    char * name;

    if(!PyArg_ParseTuple(args,"s:pvm_unexport",&name)) 
        return NULL;
    pvm_unexport(name);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *pypvm_unpack(PyObject *self, PyObject *args) {
  /* Obviously,  I have to do something with this!!!!!!!!! */
  return_none(PvmNotImpl);
}

static PyObject *pypvm_upklong(PyObject *self, PyObject *args) {
  PyObject * resultobj,*currentobj;
  int nitem, stride = 1;
  long * data;
  int info;
  int index;

  if (!PyArg_ParseTuple(args,"i|i:pvm_upkint",&nitem,&stride)) return NULL;
  data = (long *) calloc(nitem,sizeof(long));
  info = pvm_upklong(data,nitem,stride);
  if (was_error(info)) { free(data); return NULL; }
  resultobj = PyList_New(nitem);
  for (index=0;index<nitem;index++) {
    currentobj = PyInt_FromLong(data[index]);
    PyList_SetItem(resultobj,index,currentobj);
  }
  free(data);
  return resultobj;  
}


static PyObject *pypvm_upkdouble(PyObject *self, PyObject *args) {
  PyObject * resultobj,*currentobj;
  int nitem, stride = 1;
  double * data;
  int info;
  int index;

  if (!PyArg_ParseTuple(args,"i|i:pvm_upkint",&nitem,&stride)) return NULL;
  data = (double *) calloc(nitem,sizeof(double));
  info = pvm_upkdouble(data,nitem,stride);
  if (was_error(info)) { free(data); return NULL; }
  resultobj = PyList_New(nitem);
  for (index=0;index<nitem;index++) {
    currentobj = PyFloat_FromDouble(data[index]);
    PyList_SetItem(resultobj,index,currentobj);
  }
  free(data);
  return resultobj;  
}



static PyObject *pypvm_upkstr(PyObject *self,PyObject *args) {
  PyObject * resultobj;
  int size=100;
  int info;
  char * data;
  
  if (!PyArg_ParseTuple(args,"|i",&size)) return NULL;
  data = (char *) malloc (size+1);
  info=pvm_upkstr(data);
  if (was_error(info)) { free(data); return NULL; }
  resultobj = PyString_FromString(data);
  free(data);
  return resultobj;
}

static PyObject *pypvm_version(PyObject *self, PyObject *args) {
    PyObject * resultobj;
    char * result;

    result = pvm_version();
    resultobj = Py_BuildValue("s", result);
    return resultobj;
}

static PyMethodDef pvmMethods[] = {
	 { "version", pypvm_version, 1 },
	 { "unexport", pypvm_unexport, 1 },
	 /*	 { "trecv", pypvm_trecv, 1 }, */
	 { "tidtohost", pypvm_tidtohost, 1 },
	 /*{ "start_pvmd", pypvm_start_pvmd, 1 }, */
	 { "spawn", pypvm_spawn, 1 }, 
	 { "setsbuf", pypvm_setsbuf, 1 },
	 { "setrbuf", pypvm_setrbuf, 1 },
	 /* { "setopt", pypvm_setopt, 1 }, */
	 { "setmwid", pypvm_setmwid, 1 },
	 { "sendsig", pypvm_sendsig, 1 },
	 { "send", pypvm_send, 1 },
	 { "register_tasker", pypvm_reg_tasker, 1 },
	 { "recv", pypvm_recv, 1 },
	 { "pstat", pypvm_pstat, 1 },
	 { "probe", pypvm_probe, 1 },
	 { "perror", pypvm_perror, 1 },
	 { "parent", pypvm_parent, 1 },
	 { "nrecv", pypvm_nrecv, 1 },
	 { "mytid", pypvm_mytid, 1 },
	 { "mstat", pypvm_mstat, 1 },
	 { "mkbuf", pypvm_mkbuf, 1 },
	 /* 	 { "mcast", pypvm_mcast, 1 }, */
	 { "lvgroup", pypvm_lvgroup, 1 },
	 { "lookup", pypvm_lookup, 1 },
	 { "kill", pypvm_kill, 1 },
	 { "joingroup", pypvm_joingroup, 1 },
	 { "insert", pypvm_insert, 1 },
	 { "initsend", pypvm_initsend, 1 },
	 { "halt", pypvm_halt, 1 },
	 { "gsize", pypvm_gsize, 1 },
	 { "gettid", pypvm_gettid, 1 },
	 { "getsbuf", pypvm_getsbuf, 1 },
	 { "getrbuf", pypvm_getrbuf, 1 },
	 /* { "getopt", pypvm_getopt, 1 }, */
	 { "getmwid", pypvm_getmwid, 1 },
	 { "getinst", pypvm_getinst, 1 },
	 /* { "gather", pypvm_gather, 1}, */
	 /* { "getfds", pypvm_getfds, 1 }, */
	 { "freebuf", pypvm_freebuf, 1 },
	 { "export", pypvm_export, 1 },
	 { "exit", pypvm_exit, 1 },
	 /* { "delhosts", pypvm_delhosts, 1 }, */
	 { "delete", pypvm_delete, 1 },
	 { "tasks", pypvm_tasks, 1 },
	 { "configuration", pypvm_hostinfo, 1 },
	 { "bufinfo", pypvm_bufinfo, 1 },
	 { "bcast", pypvm_bcast, 1 },
	 { "barrier", pypvm_barrier, 1 },
	 { "archcode", pypvm_archcode, 1 },
	 { "pack", pypvm_pack, 1 },
	 { "upklong", pypvm_upklong, 1 },
	 { "upkdouble", pypvm_upkdouble, 1 },
	 { "upkstr", pypvm_upkstr, 1 },
	 { NULL, NULL }
};


void initpvm() 
{
  PyObject * exceptionDictionary, exceptionModule;
  PyObject * dataDictionary, dataModule;
  PyObject * spawnDictionary, spawnModule;
  PyObject * notifyDictionary, notifyModule;

  pvm_setopt(PvmAutoErr,0);
  /* make things quiet.  We raise an excpetion on every error anyway,  so
     it provides no extra error reporting than what we have already. */

  pypvm_module = Py_InitModule("pvm", pvmMethods);
  pypvm_dictionary = PyModule_GetDict(pypvm_module);
  
  /* exceptionModule = Py_InitModule("exception",pvmExceptionMethods); */
  exceptionDictionary = PyDict_New();
  dataDictionary = PyDict_New();
  spawnDictionary = PyDict_New();
  notifyDictionary = PyDict_New();
  
  /* I am wondering whether I should remove the clutter in the
     module name space by having Data and Task and ... dictionaries
     instead. */
  PyDict_SetItemString(pypvm_dictionary,"data",dataDictionary);
  PyDict_SetItemString(pypvm_dictionary,"spawnOpts",spawnDictionary);
  PyDict_SetItemString(pypvm_dictionary,"exception",exceptionDictionary);
  PyDict_SetItemString(pypvm_dictionary,"notify",notifyDictionary);
  
  /* Data definitions for pvm_initsend and pvm_initbuf */
  PyDict_SetItemString(dataDictionary,"default",PyInt_FromLong((long) PvmDataDefault));
  PyDict_SetItemString(dataDictionary,"raw",    PyInt_FromLong((long) PvmDataRaw));
  /* The following isn't possible,  I suspect... */
  /*
  PyDict_SetItemString(dataDictionary,"inplace", PyInt_FromLong((long) PvmDataInPlace));
  PyDict_SetItemString(pypvm_dictionary,"DataFoo",     PyInt_FromLong((long) PvmDataFoo));
  */
  
  /* Flags for pvm_spawn and such like */
  PyDict_SetItemString(spawnDictionary,"TaskDefault", PyInt_FromLong((long) PvmTaskDefault));
  PyDict_SetItemString(spawnDictionary,"TaskHost",    PyInt_FromLong((long) PvmTaskHost));
  PyDict_SetItemString(spawnDictionary,"TaskArch",    PyInt_FromLong((long) PvmTaskArch));
  PyDict_SetItemString(spawnDictionary,"TaskDebug",   PyInt_FromLong((long) PvmTaskDebug));
  PyDict_SetItemString(spawnDictionary,"TaskTrace",   PyInt_FromLong((long) PvmTaskTrace));
  PyDict_SetItemString(spawnDictionary,"MppFront",    PyInt_FromLong((long) PvmMppFront));
  PyDict_SetItemString(spawnDictionary,"HostCompl",   PyInt_FromLong((long) PvmHostCompl));
  
  /* for pvm_notify */
  PyDict_SetItemString(notifyDictionary,"TaskExit", PyInt_FromLong((long) PvmTaskExit));
  PyDict_SetItemString(notifyDictionary,"HostDelete", PyInt_FromLong((long) PvmHostDelete));
  PyDict_SetItemString(notifyDictionary,"HostAdd",  PyInt_FromLong((long) PvmHostAdd));
  
  /* for pvm_setopt and pvm_getopt */
  /* I _really_ will wrap these as a dictionary, somehow */
  PyDict_SetItemString(pypvm_dictionary,"PvmRoute", PyInt_FromLong((long) PvmRoute));
    /* PvmRoute options */
    PyDict_SetItemString(pypvm_dictionary,"PvmDontRoute", PyInt_FromLong((long) PvmDontRoute));
    PyDict_SetItemString(pypvm_dictionary,"PvmAllowDirect",PyInt_FromLong((long) PvmAllowDirect));
    PyDict_SetItemString(pypvm_dictionary,"PvmRouteDirect",PyInt_FromLong((long) PvmRouteDirect));
	   
    /*
  PyDict_SetItemString(pypvm_dictionary,"PvmDebugMask", PyInt_FromLong((long) PvmDebugMask));
  PyDict_SetItemString(pypvm_dictionary,"PvmAutoErr",   PyInt_FromLong((long) PvmAutoErr));
  PyDict_SetItemString(pypvm_dictionary,"PvmOutputTid", PyInt_FromLong((long) PvmOutputTid));
  PyDict_SetItemString(pypvm_dictionary,"PvmOutputCode",PyInt_FromLong((long) PvmOutputCode));
  PyDict_SetItemString(pypvm_dictionary,"PvmTraceTid",  PyInt_FromLong((long) PvmTraceTid));
  PyDict_SetItemString(pypvm_dictionary,"PvmTraceCode", PyInt_FromLong((long) PvmTraceCode));
  PyDict_SetItemString(pypvm_dictionary,"PvmFragSize",  PyInt_FromLong((long) PvmFragSize));
  PyDict_SetItemString(pypvm_dictionary,"PvmResvTids",  PyInt_FromLong((long) PvmResvTids));
  PyDict_SetItemString(pypvm_dictionary,"PvmSelfOutputTid", PyInt_FromLong((long) PvmSelfOutputTid));
  PyDict_SetItemString(pypvm_dictionary,"PvmSelfOutputCode", PyInt_FromLong((long) PvmSelfOutputCode));	 
  PyDict_SetItemString(pypvm_dictionary,"PvmSelfTraceTid",PyInt_FromLong((long) PvmSelfTraceTid));
  PyDict_SetItemString(pypvm_dictionary,"PvmSelfTraceCode",PyInt_FromLong((long) PvmSelfTraceCode));
  PyDict_SetItemString(pypvm_dictionary,"PvmShowTids", PyInt_FromLong((long) PvmShowTids));
  PyDict_SetItemString(pypvm_dictionary,"PvmPollType", PyInt_FromLong((long) PvmPollType));
  PyDict_SetItemString(pypvm_dictionary,"PvmPollConstant",PyInt_FromLong((long) PvmPollConstant));
  PyDict_SetItemString(pypvm_dictionary,"PvmPollSleep", PyInt_FromLong((long) PvmPollSleep));
  PyDict_SetItemString(pypvm_dictionary,"PvmPollTime", PyInt_FromLong((long) PvmPollTime));
  PyDict_SetItemString(pypvm_dictionary,"PvmTaskSelf", PyInt_FromLong((long) PvmTaskSelf));
  PyDict_SetItemString(pypvm_dictionary,"PvmTaskChild", PyInt_FromLong((long) PvmTaskChild));
  */
	 /* Error status values.  These will often be raised as exceptions */

  
  PyDict_SetItemString(pypvm_dictionary,"Ok", PyInt_FromLong((long) PvmOk));
  PyDict_SetItemString(exceptionDictionary,"Ok",PyInt_FromLong((long) PvmOk));
  
  
#define DefineException(x)  \
  x ## Exception = PyString_FromString(#x); \
  PyDict_SetItemString(exceptionDictionary, #x "Str",x ## Exception);  \
  x ## Number = PyInt_FromLong((long) x); \
  PyDict_SetItemString(exceptionDictionary, #x , x ## Number);  

  DefineException(PvmBadParam); 
  DefineException(PvmMismatch);	 
  DefineException(PvmOverflow);	 
  DefineException(PvmNoData );	 
  DefineException(PvmNoHost);	 
  DefineException(PvmNoFile);	 
  DefineException(PvmNoMem );	 
  DefineException(PvmBadMsg);	 
  DefineException(PvmSysErr);	 
  DefineException(PvmNoBuf );	 
  DefineException(PvmNoSuchBuf); 
  DefineException(PvmNullGroup);
  DefineException(PvmDupGroup); 
  DefineException(PvmNoGroup);   
  DefineException(PvmNotInGroup );
  DefineException(PvmNoInst );	 
  DefineException(PvmHostFail);  
  DefineException(PvmNoParent );	 
  DefineException(PvmNotImpl ); 
  DefineException(PvmDSysErr);	 
  DefineException(PvmBadVersion);	 
  DefineException(PvmOutOfRes); 
  DefineException(PvmDupHost);	 
  DefineException(PvmCantStart);	 
  DefineException(PvmAlready ); 
  DefineException(PvmNoTask);	 
  DefineException(PvmNoEntry);	 
  DefineException(PvmDupEntry);

/*


  PvmBadParamException = PyInt_FromLong((long) PvmBadParam);
  PyDict_SetItemString(pypvm_dictionary,"PvmBadParam",PvmBadParamException); 

  PvmMismatchException = PyInt_FromLong((long) PvmMismatch);
  PyDict_SetItemString(pypvm_dictionary,"PvmMismatch", PvmMismatchException);

  PvmOverflowException = PyInt_FromLong((long)	 PvmOverflow);
  PyDict_SetItemString(pypvm_dictionary,"PvmOverflow", 	 PvmOverflowException);

  PvmNoDataException = PyInt_FromLong((long)	 PvmNoData);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoData", 	 PvmNoDataException);

  PvmNoHostException = PyInt_FromLong((long)	 PvmNoHost);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoHost",  PvmNoHostException);

  PvmNoFileException = PyInt_FromLong((long)	 PvmNoFile);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoFile", 	 PvmNoFileException);
  
  PvmNoMemException = PyInt_FromLong((long)	 PvmNoMem);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoMem", 	 PvmNoMemException);
  
  PvmBadMsgException =PyInt_FromLong((long)	 PvmBadMsg);
  PyDict_SetItemString(pypvm_dictionary,"PvmBadMsg", 	 PvmBadMsgException);
  
  PvmSysErrException =PyInt_FromLong((long)	 PvmSysErr);
  PyDict_SetItemString(pypvm_dictionary,"PvmSysErr", 	 PvmSysErrException);
  
  PvmNoBufException = PyInt_FromLong((long)	 PvmNoBuf);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoBuf", 	 PvmNoBufException);
  
  PvmNoSuchBufException = PyInt_FromLong((long)	 PvmNoSuchBuf);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoSuchBuf", 	 PvmNoSuchBufException);
  
  PvmNullGroupException = PyInt_FromLong((long)	 PvmNullGroup);
  PyDict_SetItemString(pypvm_dictionary,"PvmNullGroup", 	 PvmNullGroupException);
  
  PvmDupGroupException = PyInt_FromLong((long)	 PvmDupGroup);
  PyDict_SetItemString(pypvm_dictionary,"PvmDupGroup", 	 PvmDupGroupException);
  
  PvmNoGroupException = PyInt_FromLong((long)	 PvmNoGroup);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoGroup", 	 PvmNoGroupException);
  
  PvmNotInGroupException = PyInt_FromLong((long)	 PvmNotInGroup);
  PyDict_SetItemString(pypvm_dictionary,"PvmNotInGroup",  PvmNotInGroupException);
  
  PvmNoInstException = PyInt_FromLong((long)	 PvmNoInst);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoInst", 	 PvmNoInstException);
  
  PvmHostFailException = PyInt_FromLong((long)	 PvmHostFail);
  PyDict_SetItemString(pypvm_dictionary,"PvmHostFail", 	 PvmHostFailException);
  
  PvmNoParentException = PyInt_FromLong((long)	 PvmNoParent);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoParent", 	 PvmNoParentException);
  
  PvmNotImplException = PyInt_FromLong((long)	 PvmNotImpl);
  PyDict_SetItemString(pypvm_dictionary,"PvmNotImpl", 	 PvmNotImplException);
  
  PvmDSysErrException = PyInt_FromLong((long)	 PvmDSysErr);
  PyDict_SetItemString(pypvm_dictionary,"PvmDSysErr", 	 PvmDSysErrException);
  
  PvmBadVersionException = PyInt_FromLong((long)	 PvmBadVersion);
  PyDict_SetItemString(pypvm_dictionary,"PvmBadVersion",  PvmBadVersionException);
  
  PvmOutOfResException = PyInt_FromLong((long)	 PvmOutOfRes);
  PyDict_SetItemString(pypvm_dictionary,"PvmOutOfRes", 	 PvmOutOfResException);
  
  PvmDupHostException = PyInt_FromLong((long)	 PvmDupHost);
  PyDict_SetItemString(pypvm_dictionary,"PvmDupHost", 	 PvmDupHostException);
  
  PvmCantStartException = PyInt_FromLong((long)	 PvmCantStart);
  PyDict_SetItemString(pypvm_dictionary,"PvmCantStart", 	 PvmCantStartException);
  
  PvmAlreadyException = PyInt_FromLong((long)	 PvmAlready);
  PyDict_SetItemString(pypvm_dictionary,"PvmAlready", 	 PvmAlreadyException);
  
  PvmNoTaskException = PyInt_FromLong((long)	 PvmNoTask);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoTask", 	 PvmNoTaskException);
  
  PvmNoEntryException =PyInt_FromLong((long)	 PvmNoEntry);
  PyDict_SetItemString(pypvm_dictionary,"PvmNoEntry", 	 PvmNoEntryException);
  
  PvmDupEntryException = PyInt_FromLong((long)	 PvmDupEntry);
  PyDict_SetItemString(pypvm_dictionary,"PvmDupEntry", 	 PvmDupEntryException);
  */
  
  /* For pvm_gather.  I don't think they'll be necessary. */
  /*
  PyDict_SetItemString(pypvm_dictionary,"PVM_STR", PyInt_FromLong((long) PVM_STR));
  PyDict_SetItemString(pypvm_dictionary,"PVM_BYTE", PyInt_FromLong((long) PVM_BYTE));
  PyDict_SetItemString(pypvm_dictionary,"PVM_SHORT", PyInt_FromLong((long) PVM_SHORT));
  PyDict_SetItemString(pypvm_dictionary,"PVM_INT", PyInt_FromLong((long) PVM_INT));
  PyDict_SetItemString(pypvm_dictionary,"PVM_FLOAT", PyInt_FromLong((long) PVM_FLOAT));
  PyDict_SetItemString(pypvm_dictionary,"PVM_CPLX", PyInt_FromLong((long) PVM_CPLX));
  PyDict_SetItemString(pypvm_dictionary,"PVM_DOUBLE", PyInt_FromLong((long)PVM_DOUBLE));
  PyDict_SetItemString(pypvm_dictionary,"PVM_DCPLX", PyInt_FromLong((long) PVM_DCPLX));
  PyDict_SetItemString(pypvm_dictionary,"PVM_LONG", PyInt_FromLong((long) PVM_LONG));
  PyDict_SetItemString(pypvm_dictionary,"PVM_USHORT", PyInt_FromLong((long)PVM_USHORT));
  PyDict_SetItemString(pypvm_dictionary,"PVM_UINT", PyInt_FromLong((long) PVM_UINT));
  PyDict_SetItemString(pypvm_dictionary,"PVM_ULONG", PyInt_FromLong((long) PVM_ULONG));
  */
}
